from OpenGL.error import GLError
import OpenGL.GL as gl
import os
import numpy as np

#==============================================================================
class Shader(object):
    """ Base shader class
    """
    def __init__(self, path=None, **kargs):

        self._uniforms = {}
        self._setters = {}
        self._sourceCode = {}
        # create the program handle
        self.program = gl.glCreateProgram()
        if not self.program:
            raise SystemExit("Shader creation failed: "
                    "Could not find valid memory location in constructor")
        # we are not linked yet
        self._linked = False

        for stype in kargs:
            shader = kargs[stype]
            if path is not None:
                shader = os.path.join(path, shader)
            if shader and os.path.isfile(shader):
                with open(shader) as f:
                    self._sourceCode[stype] = f.read()
            else:
                self._sourceCode[stype] = f.read()

            code = self._sourceCode[stype]
            if code:
                if stype == 'vertex':
                    self._build_shader(code, gl.GL_VERTEX_SHADER)
                elif stype == 'fragment':
                    self._build_shader(code, gl.GL_FRAGMENT_SHADER)
                elif stype == 'geometry':
                    self._build_shader(code, gl.GL_GEOMETRY_SHADER)
                elif stype == 'compute':
                    self._build_shader(code, gl.GL_COMPUTE_SHADER)
                else:
                    raise SystemExit("Unimplemented shader type '%s': "
                    "Giving up...")

        self._link()

    def addUniform(self, uniform, utype=None):
        """ Add a uniform variable to the shader.
            If utype is not none it defines the setter function.
            Valid utypes are i, b, f, 2f, 4fv
        """
        uniformLocation = gl.glGetUniformLocation(self.program, uniform)
        if uniformLocation == -1:
            print('Cannot find uniform %s' % uniform)
            return
        self._uniforms[uniform] = uniformLocation
        if utype is not None:
            if utype in "i b".split():
                self._setters[uniform] = gl.glUniform1i
            elif utype == 'f':
                self._setters[uniform] = gl.glUniform1f
            elif utype == '2f':
                self._setters[uniform] = gl.glUniform2f
            elif utype == 'mat4':
                self._setters[uniform] = gl.glUniformMatrix4fv
            else:
                raise SystemExit("Unimplemented utype '%s'" % utype)

    def addUniforms(self, uniformlist):
        for uniform in uniformlist:
            if isinstance(uniform, tuple):
                self.addUniform(*uniform)
            else:
                self.addUniform(uniform)

    def _build_shader(self, sourceCode, shader_type):
        """ Actual building of the shader
        """
        # create the shader handle
        shader = gl.glCreateShader(shader_type)
        if shader == 0:
            raise SystemExit("Shader creation failed: "
                    "Could not find valid memory location when adding shader")

        # Upload shader code
        gl.glShaderSource(shader, sourceCode)

        try:
            # compile the shader
            gl.glCompileShader(shader)
        except GLError as e:
            raise SystemExit(gl.glGetShaderInfoLog(shader))
        else:
            gl.glAttachShader(self.program, shader)

    def delete(self):
        programs = gl.glGetAttachedShaders(self.program)
        for item in programs:
            gl.glDetachShader(self.program, item)
            gl.glDeleteShader(item)
        gl.glDeleteProgram(self.program)

    def _link(self):
        """ Link the program
        """
        gl.glLinkProgram(self.program)
        # retrieve the link status

        if not gl.glGetProgramiv(self.program, gl.GL_LINK_STATUS):
            raise SystemExit(gl.glGetProgramInfoLog(self.program))
        else:
            gl.glValidateProgram(self.program)
            self._linked = True

        if not gl.glGetProgramiv(self.program, gl.GL_VALIDATE_STATUS):
            raise SystemExit(gl.glGetProgramInfoLog(self.program))

    def bind(self):
        """ Bind the program, i.e. use it
        """
        gl.glUseProgram(self.program)

    def unbind(self):
        """ Unbind whatever program is currently bound - not necessarily this
            program, so this should probably be a class method instead.
        """
        gl.glUseProgram(0)

    def setUniform(self, name, value):
        if isinstance(value, (list, tuple)):
            self._setters[name](self._uniforms.get(name), *value)
        elif isinstance(value, np.ndarray) and value.shape == (4, 4):
            self._setters[name](self._uniforms.get(name), 1, True, value)
        else:
            try:
                self._setters[name](self._uniforms.get(name), value)
            except:
                raise SystemExit("Wrong setter function for '%s'" % name)

    def setUniforms(self, nameandvalue):
        for name, value in nameandvalue:
            self.setUniform(name, value)



import os
import numpy as np
import ctypes
import OpenGL.GL as gl

# Local imports
from .shader import Shader
from .texture import Texture
from .__version__ import __version__

# Fix ramdom sequence seed
np.random.seed(123)

GLSLDIR = os.path.join(os.path.dirname(__file__ ), 'glsl')
WORKGROUP_SIZE = 32

#------------------------------------------------------------------------------
def glInfo():
    """ Return OpenGL information dict
        WARNING: OpenGL context MUST be initialized !!!

        Args:
            None

        Returns:
            OpenGL information dict
    """
    major = gl.glGetIntegerv(gl.GL_MAJOR_VERSION)
    minor = gl.glGetIntegerv(gl.GL_MINOR_VERSION)
    version = gl.glGetString(gl.GL_VERSION)
    vendor = gl.glGetString(gl.GL_VENDOR)
    renderer = gl.glGetString(gl.GL_RENDERER)
    glsl = gl.glGetString(gl.GL_SHADING_LANGUAGE_VERSION)
    glversion = float("%d.%d" % (major, minor))
    retval = {
            'glversion': glversion,
            'version': version,
            'vendor': vendor,
            'renderer': renderer,
            'glsl': glsl,
            }
    if glversion >= 4.3:
        count = np.zeros(3, dtype=np.int32)
        size = np.zeros(3, dtype=np.int32)
        count[0] = gl.glGetIntegeri_v(gl.GL_MAX_COMPUTE_WORK_GROUP_COUNT, 0)[0]
        count[1] = gl.glGetIntegeri_v(gl.GL_MAX_COMPUTE_WORK_GROUP_COUNT, 1)[0]
        count[2] = gl.glGetIntegeri_v(gl.GL_MAX_COMPUTE_WORK_GROUP_COUNT, 2)[0]
        size[0] = gl.glGetIntegeri_v(gl.GL_MAX_COMPUTE_WORK_GROUP_SIZE, 0)[0]
        size[1] = gl.glGetIntegeri_v(gl.GL_MAX_COMPUTE_WORK_GROUP_SIZE, 1)[0]
        size[2] = gl.glGetIntegeri_v(gl.GL_MAX_COMPUTE_WORK_GROUP_SIZE, 2)[0]
        retval['maxComputeWorkGroupCount'] = count
        retval['maxComputeWorkGroupSize'] = size
    return retval

#------------------------------------------------------------------------------
def field2RGB(field):
    """ Return 2D field converted to uint8 RGB image (i.e. scaled in [0, 255])

        Args:
            field (:class:`numpy.ndarray`): (u, v) 2D vector field instance

        Returns:
            (
            rgb (:class:`numpy.ndarray`): uint8 RGB image,
            uMin (float): u min,
            uMax (float): u max,
            vMin (float): v min,
            vMax (float): v max,
            ) (tuple): Return value

    """
    rows = field.shape[0]
    cols = field.shape[1]
    u = field[:, :, 0]
    v = field[:, :, 1]
    uMin = u.min()
    uMax = u.max()
    vMin = v.min()
    vMax = v.max()
    rgb = np.zeros((rows, cols, 3), dtype=np.uint8)
    rgb[:, :, 0] = 255 * (u - uMin) / ( uMax - uMin)
    rgb[:, :, 1] = 255 * (v - vMin) / ( vMax - vMin)
    return rgb, uMin, uMax, vMin, vMax

#------------------------------------------------------------------------------
def modulus(field):
    """ Return normalized modulus of 2D field image

        Returns:
            normalized modulus of 2D field image (i.e. scaled in [0, 1.])
    """
    modulus = np.flipud(np.hypot(field[:, :, 0], field[:,:,1]))
    return np.asarray(modulus/modulus.max(), np.float32)

#==============================================================================
class FieldAnimation(object):
    """ Field Animation with OpenGL


        1. draw the modulus of the vector field or a user defined image
            if requested;
        2. set a framebuffer texture (screen texture) as the main
            rendering target:
                (a) draw the background texture on the screen texture
                    with a fixed opacity;
                (b) decode the particles positions from the
                    currentTracersPosition texture and draw them on
                    the screen texture;
        3. set the rendering target to the active window;
        4. draw screen texture on the active window;
        5. swap screen texture and background texture;
        6. calculate the new particles positions
            (in the update shader) and encode them in the
            nextTracersPosition texture;
        7. swap nextTracersPosition texture and
            currentTracersPosition texture;
    """
    def __init__(self, width, height, field, computeSahder=False,
            image=None):
        """ Animate 2D vector field

            Args:
                width (int): width in pixels
                height (int): height in pixels
                field (np.ndarray): 2D vector field
                cs = True selects the compute shader version
                image = Optional background image
        """
        self.useComputeShader = computeSahder
        self.image = image
        # Parameters that can be changed later
        self.periodic = True
        self.drawField = False
        self.fadeOpacity = 0.996
        self.decayBoost = 0.01
        self.speedFactor = 0.25
        self.decay = 0.003
        self.palette = True
        self.color = (0.5, 1.0, 1.0)
        self.pointSize = 1.0
        self._tracersCount = 10000
        self.fieldScaling = 1.0

        # These are fixed
        self.w_width = width
        self.w_height = height

        # Since points are in [0, 1] a traslation and a scaling is needed on
        # the model matrix
        T = np.eye(4, dtype=np.float32)
        T[:, -1] = (-1., 1., 0, 1)
        S = (np.eye(4, dtype=np.float32)
                * np.array((2., -2., 1., 1.), dtype=np.float32))
        # Model transform matrix
        model = np.dot(T, S)
        # View matrix
        view = np.eye(4)

        # Projection matrix
        proj = np.eye(4)
        self.drawMVP = np.dot(model, np.dot(view, proj))
        self.fieldMVP = np.eye(4)

        # CubeHelix color palette parameters
        cubeHelixParams =(
                ('start', 'f'),
                ('gamma', 'f'),
                ('rot', 'f'),
                ('reverse', 'b'),
                ('minSat', 'f'),
                ('maxSat', 'f'),
                ('minLight', 'f'),
                ('maxLight', 'f'),
                ('startHue', 'f'),
                ('endHue', 'f'),
                ('useHue', 'b'),
                )

        # Create Shader program and uniforms for the vector field
        self.fieldProgram = Shader(vertex='field.vert', fragment='field.frag',
                path=GLSLDIR)
        self.fieldProgram.addUniforms((('gMap', 'i'),
                ('MVP', 'mat4'),
                ) + cubeHelixParams)

        # Create Shader program and uniforms for the tracers
        self.drawProgram = Shader(vertex='draw.vert', fragment='draw.frag',
                path=GLSLDIR)
        self.drawProgram.addUniforms((
            ('MVP', 'mat4'),
            ('u_tracers', 'i'),
            ('u_tracersRes', 'f'),
            ('palette', 'b'),
            ('pointSize', 'f'),
            ('u_field', 'i'),
            ('u_fieldMin', '2f'),
            ('u_fieldMax', '2f')) + cubeHelixParams)

        # Create Shader program and uniforms for updating the screen
        self.screenProgram = Shader(vertex='quad.vert',
                fragment='screen.frag', path=GLSLDIR)
        self.screenProgram.addUniforms((
            ('u_screen', 'i'),
            ('u_opacity', 'f')))

        # Create image background shader
        if self.image is not None:
            self.imageProgram = Shader(vertex='field.vert',
                    fragment='image.frag', path=GLSLDIR)
            self.imageProgram.addUniforms((('gMap', 'i'),
                ('MVP', 'mat4'),))

        # Create Shader program and uniforms for updating the tracers position
        if self.useComputeShader:
            self.updateProgram = Shader(compute='update.comp', path=GLSLDIR)
        else:
            self.updateProgram = Shader(vertex='quad.vert',
                    fragment='update.frag', path=GLSLDIR)
        self.updateProgram.addUniforms((
                ('u_tracers', 'i'),
                ('u_field', 'i'),
                ('u_fieldRes', '2f'),
                ('u_fieldMin', '2f'),
                ('u_fieldMax', '2f'),
                ('u_rand_seed', 'f'),
                ('u_speed_factor', 'f'),
                ('u_decay', 'f'),
                ('u_decay_boost', 'f'),
                ('fieldScaling', 'f'),
                ('periodic', 'b')))

        # Set the vector field
        self.setField(field)
        self._initTracers()


    def setField(self, field):
        """ Set the 2D vector field. Must be called every time a new
            vetor field is selected.

            Args:
                field (np.ndarray): 2D vector field
        """
        # Automatic field scalking
        self.fieldScaling = self.speedFactor * 0.01 / field.max()

        # Prepare the data
        self._fieldAsRGB, uMin, uMax, vMin, vMax = field2RGB(field)
        # Compute field modulus
        self.modulus = modulus(field)
        # Set values that will not change
        self.drawProgram.bind()
        self.drawProgram.setUniform('u_fieldMin', (uMin, vMin))
        self.drawProgram.setUniform('u_fieldMax', (uMax, vMax))
        self.drawProgram.unbind()
        # Set values that will not change
        self.updateProgram.bind()
        self.updateProgram.setUniform('u_fieldMin', (uMin, vMin))
        self.updateProgram.setUniform('u_fieldMax', (uMax, vMax))
        self.updateProgram.unbind()
        self._initTracers()

    def setRenderingTarget(self, texture):
        """ Set texture as rendering target

            Args:
                texture (class:`texture` instance): 2D vector field
        """
        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, self.frameBuffer)
        gl.glFramebufferTexture2D(gl.GL_FRAMEBUFFER, gl.GL_COLOR_ATTACHMENT0,
            gl.GL_TEXTURE_2D, texture.handle(), 0)

    def setSize(self, width, height):
        """ Set instance size. Must be called when the window is resized.

            Args:
                width (int): window width in pixels
                height (int): window height in pixels
        """
        self.w_width = width
        self.w_height = height
        self._initTracers()

    def resetRenderingTarget(self):
        """ Bind first (default) framebuffer and reset the viewport.
        """
        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, 0)
        gl.glViewport(0, 0, self.w_width, self.w_height)

    @property
    def tracersCount(self):
        """ Return tracers count

            Returns:
                number of tracers
        """
        return  self._tracersCount

    @tracersCount.setter
    def tracersCount(self, value):
        """ Tracer count setter method, calls self._initTracers under
            the hood.

            Args:
                value (int): number of tracers to create
        """
        self._tracersCount = value
        self._initTracers()

    def _initTracers(self):
        """ Initialize the tracers positions
        """

        # Create a buffer for the tracers
        self.emptyPixels = np.zeros((self.w_width * self.w_height * 4),
                np.uint8)

        # Initial random tracers position
        self.tracers =  np.asarray(255.0 * np.random.random(
                self._tracersCount * 4), dtype=np.uint8, order='C')
        self.tracersRes = np.ceil(np.sqrt(self.tracers.size / 4))
        self.iTracers = np.arange(self.tracers.size / 4, dtype=np.float32)

        # Create all textures
        # Tracers position stored in texture 0
        self._currentTracersPos = Texture(
                data=self.tracers,
                width=self.tracersRes, height=self.tracersRes)
        # Initial random tracers position stored in texture 1
        self._nextTracersPos = Texture(
                data=self.tracers,
                width=self.tracersRes, height=self.tracersRes)
        self.fieldTexture = Texture(data=self._fieldAsRGB,
                width=self._fieldAsRGB.shape[1],
                height=self._fieldAsRGB.shape[0],
                filt=gl.GL_LINEAR)
        self.backgroundTexture = Texture(data=self.emptyPixels,
                width=self.w_width, height=self.w_height)
        self.screenTexture = Texture(data=self.emptyPixels,
                width=self.w_width, height=self.w_height)

        # VAOS
        values = np.zeros(int(self.tracers.size/4), [('a_index', 'f4')])
        values['a_index'] = np.asarray(self.iTracers,
                dtype=np.float32, order='C')

        if self.image is not None:
            self.imageTexture = Texture(data=self.image,
                    dtype=gl.GL_UNSIGNED_BYTE)
        self.modulusTexture = Texture(data=self.modulus, dtype=gl.GL_FLOAT)

        ## VAO index
        self._vao = gl.glGenVertexArrays(1)
        vbo = gl.glGenBuffers(1)
        gl.glBindVertexArray(self._vao)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, vbo)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, values, gl.GL_STATIC_DRAW)
        gl.glEnableVertexAttribArray(0)
        gl.glVertexAttribPointer(0, 1, gl.GL_FLOAT, gl.GL_FALSE, 0,
                ctypes.c_void_p(0))
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)
        gl.glBindVertexArray(0)

        # Screen quad
        quad = np.zeros(4, dtype=[('vert', 'f4', 2),('tex', 'f4', 2)])

        quad['vert'] = np.array([
                [-1,  1],
                [1,   1],
                [1,  -1],
                [-1, -1]], np.float32)

        quad['tex'] = np.array([
                [0, 1],
                [1, 1],
                [1, 0],
                [0, 0]], np.float32)

        indices = np.array([0, 1, 2, 2, 3, 0], np.int32)

        self._vaoQuad = gl.glGenVertexArrays(1)
        gl.glBindVertexArray(self._vaoQuad)
        quadVBO = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, quadVBO)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, quad, gl.GL_STATIC_DRAW)

        self.IBO = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.IBO)
        gl.glBufferData(gl.GL_ELEMENT_ARRAY_BUFFER, indices, gl.GL_STATIC_DRAW)

        gl.glEnableVertexAttribArray(0)
        gl.glVertexAttribPointer(0, 2,   gl.GL_FLOAT, gl.GL_FALSE, 4 * 4,
                ctypes.c_void_p(0))
        gl.glEnableVertexAttribArray(1)
        gl.glVertexAttribPointer(1, 2,   gl.GL_FLOAT, gl.GL_FALSE, 4 * 4,
                ctypes.c_void_p(8))
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)
        gl.glBindVertexArray(0)

        self.frameBuffer = gl.glGenFramebuffers(1)

    def draw(self):
        """ Render the OpenGL scene. This method is called automatically
            when the scene has to be rendered and is responsible for the
            animation.
        """
        if self.drawField:
            self.drawModulus(1.0)

        if self.image is not None:
            self.drawImage()

        self.fieldTexture.bind(0)
        ## Bind texture with random tracers position
        self._currentTracersPos.bind(1)
        self.drawScreen()
        if self.useComputeShader:
            self.updateTracersCS()
        else:
            self.updateTracers()

    def drawScreen(self):
        """ Draw background texture and tracers on screen framebuffer texture
        """
        # Draw background texture and tracers on screen framebuffer texture
        self.setRenderingTarget(self.screenTexture)
        gl.glViewport(0, 0, self.w_width, self.w_height)

        self.drawTexture(self.backgroundTexture, self.fadeOpacity)
        self.drawTracers()
        self.resetRenderingTarget()

        # Draw the screen framebuffer texture on the monitor window
        # Enable blending to support drawing on top of an existing background
        #(e.g. a map)
        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        self.drawTexture(self.screenTexture, 1.0)
        gl.glDisable(gl.GL_BLEND)

        self.backgroundTexture , self.screenTexture = (
                self.screenTexture, self.backgroundTexture)

    def drawModulus(self, opacity):
        """ Draw the modulus texture.

            Args:
                opacity (float): opacity (alpha) of the texture:
                    0 --> transparent
                    1 --> opaque
        """
        self.modulusTexture.bind()
        self.fieldProgram.bind()
        self.fieldProgram.setUniforms((
            ('gMap', 0),
            ('MVP', self.fieldMVP),
            ('start', 1.0),
            ('gamma', 0.9),
            ('rot', -5.3),
            ('reverse', False),
            ('minSat', 0.2),
            ('maxSat', 5.0),
            ('minLight', 0.5),
            ('maxLight', 0.9),
            ('startHue', 240.0),
            ('endHue', -300.0),
            ('useHue', True),
            ))
        gl.glBindVertexArray(self._vaoQuad)
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.IBO)

        gl.glDrawElements(gl.GL_TRIANGLES, 6, gl.GL_UNSIGNED_INT,
                          ctypes.c_void_p(0))
        self.fieldProgram.unbind()

        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)
        gl.glBindVertexArray(0)
        self.modulusTexture.unbind()

    def drawImage(self):
        """ Draw an image texture in background.
        """
        self.imageProgram.bind()
        self.imageTexture.bind()
        self.imageProgram.setUniforms((('gMap', 0),
            ('MVP', self.fieldMVP),
            ))
        gl.glBindVertexArray(self._vaoQuad)
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.IBO)

        gl.glDrawElements(gl.GL_TRIANGLES, 6, gl.GL_UNSIGNED_INT,
                          ctypes.c_void_p(0))
        self.imageProgram.unbind()

        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)
        gl.glBindVertexArray(0)
        self.imageTexture.unbind()

    def drawTexture(self, texture, opacity):
        """ Draw `texture` on the screen.

            Args:
                texture (:class:Texture): texture instance
                opacity (float): opacity (alpha) of the texture
                    0 --> transparent
                    1 --> opaque
        """
        self.screenProgram.bind()
        gl.glBindVertexArray(self._vaoQuad)
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.IBO)
        texture.bind(2)
        self.screenProgram.setUniforms((
            ('u_screen', 2),
            ('u_opacity', opacity),
            ))

        gl.glDrawElements(gl.GL_TRIANGLES, 6, gl.GL_UNSIGNED_INT,
                          ctypes.c_void_p(0))
        self.screenProgram.unbind()
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)
        gl.glBindVertexArray(0)
        texture.unbind()

    def drawTracers(self):
        """ Draw the tracers on the screen
        """
        gl.glBindVertexArray(self._vao)
        self.drawProgram.bind()

        self.drawProgram.setUniforms((
            ('u_field', 0),
            ('palette', bool(self.palette)),
            ('u_tracers', 1),
            ('MVP', self.drawMVP),
            ('pointSize', self.pointSize),
            # CubeHelix
            ('start', 1.0),
            ('gamma', 0.9),
            ('rot', -5.3),
            ('reverse', False),
            ('minSat', 0.2),
            ('maxSat', 5.0),
            ('minLight', 0.5),
            ('maxLight', 0.9),
            ('startHue', 240.0),
            ('endHue', -300.0),
            ('useHue', True),
            ('u_tracersRes', float(self.tracersRes)),
            ))
        gl.glEnable(gl.GL_VERTEX_PROGRAM_POINT_SIZE)
        gl.glDrawArrays(gl.GL_POINTS, 0, int(self.tracers.size / 4))

        self.drawProgram.unbind()
        gl.glBindVertexArray(0)

    def updateTracers(self):
        """ Update tracers position using the fragment shader provided by
            the graphic card for computing.
        """
        self.setRenderingTarget(self._nextTracersPos)
        gl.glViewport(0, 0, int(self.tracersRes), int(self.tracersRes))

        self.updateProgram.bind()

        gl.glBindVertexArray(self._vaoQuad)
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.IBO)
        self.updateProgram.setUniforms((
            ('u_field', 0),
            ('u_tracers', 1),
            ('periodic', self.periodic),
            ('u_rand_seed', np.random.random()),
            ('u_speed_factor', self.speedFactor),
            ('u_decay', self.decay),
            ('u_decay_boost', self.decayBoost),
            ('fieldScaling', self.fieldScaling),
            ('u_fieldRes', self._fieldAsRGB.shape),
            ))
        gl.glDrawElements(gl.GL_TRIANGLES, 6, gl.GL_UNSIGNED_INT,
            ctypes.c_void_p(0))
        # Replace current tracers positions with the new ones
        self._currentTracersPos = self._nextTracersPos
        self.resetRenderingTarget()

    def updateTracersCS(self):
        """ Update tracers position using the compute shader provided by
            the graphic card for computing.
        """
        self.updateProgram.bind()
        self.updateProgram.setUniforms((
            ('u_field', 0),
            ('u_tracers', 1),
            ('periodic', self.periodic),
            ('u_rand_seed', np.random.random()),
            ('u_speed_factor', self.speedFactor),
            ('u_decay', self.decay),
            ('u_decay_boost', self.decayBoost),
            ('fieldScaling', self.fieldScaling),
            ('u_fieldRes', self._fieldAsRGB.shape),
            ))

        gl.glBindImageTexture(0, self._currentTracersPos._handle, 0,
                gl.GL_FALSE, 0, gl.GL_READ_ONLY, gl.GL_RGBA8)
        gl.glBindImageTexture(1, self._nextTracersPos._handle, 0,
                gl.GL_FALSE, 0, gl.GL_WRITE_ONLY, gl.GL_RGBA8)

        gl.glDispatchCompute(int(self.tracersRes / WORKGROUP_SIZE),
                int(self.tracersRes / WORKGROUP_SIZE), 1)
        ## Lock memory access until image process ends
        gl.glMemoryBarrier(gl.GL_SHADER_IMAGE_ACCESS_BARRIER_BIT)
        self._currentTracersPos = self._nextTracersPos

        self.resetRenderingTarget()



from abc import ABCMeta, abstractmethod
import glfw
import OpenGL.GL as gl

actionMap = {glfw.PRESS: 'press',
             glfw.RELEASE: 'release',
             glfw.REPEAT: 'repeat'}

def errorCallback(error, description):
    print('Error %s, %s' % (error, description))

#==============================================================================
class glfwApp(metaclass=ABCMeta):

    KEY_G = glfw.KEY_G
    KEY_N = glfw.KEY_N
    KEY_F = glfw.KEY_F
    PRESS = glfw.PRESS
    RELEASE = glfw.RELEASE

    def __init__(self, title='', width=800, height=600, resizable=True):
        """ Create a new glfwApp instance.

            Args:
                title (string): window title
                width (integer): window width in pixels
                height (integer): window height in pixels
                resizable (bool): if True the window can be resized
        """
        self._width = width
        self._height = height
        self._title = title
        self.bg_color = (0.0, 0.0, 0.0, 0.0)

        glfw.set_error_callback(errorCallback)

        if not glfw.init():
            raise SystemExit("Error initializing GLFW")
        if resizable:
            glfw.window_hint(glfw.RESIZABLE, gl.GL_TRUE)
        else:
            glfw.window_hint(glfw.RESIZABLE, gl.GL_FALSE)
        # Create the window
        self._window = glfw.create_window(self._width, self._height,
                self._title, None, None)

        glfw.set_window_size_callback(self._window, self.onResize)

        if not self._window:
            glfw.terminate()
            raise SystemExit

        glfw.make_context_current(self._window)
        glfw.set_key_callback(self._window, self.onKeyboard)
    
    def restoreKeyCallback(self):
        glfw.set_key_callback(self._window, self.onKeyboard)
    
    @abstractmethod
    def onResize(self, window, width, height):
        """ This method must be implemened. It is called automatically when
            the window gets resized.

            Args:
                window (class:`glfw.LP__GLFWwindow` instance): window
                width (int): window width in pixels
                height (int): window height in pixels
        """
        pass

    def onKeyboard(self, window, key, scancode, action, mode):
        """ Process keybord input. This method is called automatically when
            the user interacts with the keyboard.

            Args:
                window (class:`glfw.LP__GLFWwindow` instance): window
                key (integer): the key that was pressed
                scancode (integer):
                action (integer): PRESS, RELEASE, REPEAT
                mode (integer): modifier
        """
        if key in (glfw.KEY_ESCAPE, glfw.KEY_Q):
            glfw.set_window_should_close(self._window, 1)
        glfw.poll_events()

    def window(self):
        """ Return the window instance

            Returns:
                the window instance
        """
        return self._window

    def title(self):
        """ Return window title

            Returns:
                the window title
        """
        return self._title

    def setTitle(self, title):
        """ Set window title

            Args:
                title (string): the new window title
        """
        glfw.set_window_title(self._window, title)

    def run(self):
        """ Start the application main loop
        """
        while not glfw.window_should_close(self._window):
            gl.glClearColor(*self.bg_color)
            glfw.poll_events()
            self.renderScene()
            glfw.swap_buffers(self._window)
            
        self.close()

    def close(self):
        """ Destroy the window and terminate glfw
        """
        glfw.destroy_window(self._window)
        glfw.terminate()

    def renderScene(self):
        """ Render the scene. This method is called automatically in the run
            loop
        """
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)

if __name__ == "__main__":
    app = glfwApp('glfwApp')
    app.run()




import os
import sys
import signal
import time
from pathlib import Path
import argparse
import numpy as np
import OpenGL.GL as gl
try:
    from PIL import Image
    from .glfwBackend import glfwApp
    import imgui
    from imgui.integrations.glfw import GlfwRenderer
except ImportError as e:
    raise SystemExit("%s:\nAppliaction needs extra packages to run.\n"
            "Please install Pillow, imgui and glfw." % e)
signal.signal(signal.SIGINT, signal.SIG_DFL)

# Local imports
from fieldanimation import FieldAnimation, glInfo

CHOICES = (
    'epole',
    "Duffing's equation",
    'Structurally stable system',
    'Transcritical bifurcation',
    'Reaction diffusion',
    'Positive invariant set',
    'Spiral ccw',
    'Spiral cw',
    'wind',
    'gmod',
    )

#------------------------------------------------------------------------------
def ElectricField(q, r0, x, y):
    """ Return the electric field vector E=(Ex, Ey) due to charge q at r0.
    """
    den = np.hypot(x - r0[0], y - r0[1]) ** 1.5

    return q * (x - r0[0]) / den, q * (y - r0[1]) / den

#------------------------------------------------------------------------------
def createField(eq='Spiral ccw', m=64, n=64):
# def createField(eq='Spiral ccw', m=81, n=201):
    """ Equations are taken from
        http://reference.wolfram.com/language/ref/StreamPlot.html

        Args:
            eq (string): equation name
            m (integer): number of rows
            n (integer): number of columns
    """
    mj = m * 1j
    nj = n * 1j
    Y, X = np.mgrid[-3:3:mj, -3:3:nj]

    if eq == 'epole':
        # Return n x m x 2 2D array with Electric field values
        # generated by a (2 * charges) electric charges configuration
        charges = 1
        nq = 2 ** int(charges)
        charges = []
        for i in range(nq):
            q = i % 2 * 2 - 1
            charges.append((q,
                    (np.cos(2 * np.pi * i / nq), np.sin(2 * np.pi * i / nq))))
        U, V = np.zeros((m, n)), np.zeros((m, n))
        for charge in charges:
            ex, ey = ElectricField(*charge, x=X, y=Y)
            U += ex
            V += ey

    elif eq == "Duffing's equation":
        # Duffing's equation
        U = Y.copy()
        V = X - X**3
    elif eq == 'Structurally stable system':
        # Structurally stable system
        U = 2 * X + Y - X * (X**2 + Y**2)
        V = -Y - Y * (X**2 + Y**2)
    elif eq == 'Transcritical bifurcation':
        # Transcritical bifurcation
        U = X**2
        V = -Y
    elif eq == 'Reaction diffusion':
        U = 2 * (Y - X) + X * (1 - X ** 2)
        V = -2 * (Y - X) + Y * (1 - Y ** 2)
    elif eq == 'Positive invariant set':
        U = Y.copy()
        V = - X + Y * (1 - X ** 2 - 2 * Y ** 2)
    elif eq == 'Spiral ccw':
        origin = (0.8, 0.8)
        X, Y = X - origin[0], Y - origin[1]
        U = X - Y
        V = X + Y
    elif eq == 'Spiral cw':
        origin = (0.8, 0.8)
        X, Y = X - origin[0], Y - origin[1]
        U = Y - X
        V = -X - Y
    elif eq == 'wind':
        mydir = Path(__file__).parent
        field = np.load(mydir / "wind_2016-11-20T00-00Z.npy")
        r, c, bands = field.shape
        Y, X = np.mgrid[0:r, 0:c]
        U = field[:, :, 0][::-1]
        V = - field[:, :, 1][::-1]
    elif eq == 'gmod':
        mydir = Path(__file__).parent
        U = np.load(mydir / 'vx.npy')[::-1]
        V = np.load(mydir / 'vy.npy')[::-1]
        r,c = U.shape
        Y, X = np.mgrid[0:r, 0:c]
    else:
        raise SystemExit("Unknown field. Giving up...")

    return np.flipud(np.dstack((U, -V)))

#------------------------------------------------------------------------------
def userInterface(graphicItem, app):
    """ Control graphicItem parameters interactively
    """
    imgui.new_frame()
    expanded, opened = imgui.begin('Controls', closable=True,
            flags=imgui.WINDOW_ALWAYS_AUTO_RESIZE)
    if opened:
        # Field to show
        current = app.ifield
        changed, current = imgui.combo('Field', current, list(CHOICES))
        if changed:
            app.setField(current)

        # Speed Rate
        changed, speed = imgui.drag_float('Speed',
            graphicItem.speedFactor, 0.01, 0.0, 10.0)
        if changed:
            graphicItem.speedFactor = speed

        # Decay
        changed, decay = imgui.drag_float('Decay',
            graphicItem.decay, 0.001, 0.001, 1.0)
        if changed:
            graphicItem.decay = decay

        # Drop Rate Bump
        changed, decayBoost = imgui.drag_float('Decay boost',
            graphicItem.decayBoost, 0.01, 0.001, 1.0)
        if changed:
            graphicItem.decayBoost = decayBoost

        # Unknown const
        changed, opacity = imgui.drag_float('Opacity',
            graphicItem.fadeOpacity, 0.001, 0.900, 0.999, '%.4f')
        if changed:
            graphicItem.fadeOpacity = opacity

        # Palette
        changed, color = imgui.color_edit3('Color', *graphicItem.color)
        if changed:
            graphicItem.color = color
        imgui.same_line()
        changed, palette = imgui.checkbox("Palette", graphicItem.palette)
        if changed:
            graphicItem.palette = palette

        changed, bg_color = imgui.color_edit4('Background color',
                *app.bg_color)
        if changed:
            app.bg_color = bg_color

        # Point size
        changed, pointSize = imgui.input_int("Point size",
            graphicItem.pointSize, 1, 1, 1)
        if changed:
            if pointSize > 5:
                pointSize = 5
            elif pointSize < 1:
                pointSize = 1
            graphicItem.pointSize = pointSize

        # Number of Points
        changed, tracersCount = imgui.drag_int("Number of "
            "Tracers", graphicItem.tracersCount, 1000.0, 4000, 10000000)
        if changed:
            graphicItem.tracersCount = tracersCount

        # Periodic border
        changed, periodic = imgui.checkbox("Periodic", graphicItem.periodic)
        if changed:
            graphicItem.periodic = periodic

        # Draw field
        changed, drawfield = imgui.checkbox("Draw Field",
                graphicItem.drawField)
        if changed:
            graphicItem.drawField = drawfield
    imgui.end()
    imgui.render()
    return opened

#==============================================================================
class GLApp(glfwApp):
    def __init__(self, title, width, height, options):
        super(GLApp, self).__init__(title, width, height)

        if options.image is not None:
            options.image = np.flipud(
                    np.asarray(Image.open(options.image), np.uint8))

        self.ifield = CHOICES.index(options.choose)
        field = createField(CHOICES[self.ifield])

        glversion = glInfo()['glversion']
        if not options.use_fragment and glversion < 4.3:
            print("WARNING..... Compute shaders not available with OpenGL"
                    " ver. %.1f." % glversion)
            useCompute = False
        else:
            useCompute = True
        # Add Field Animation overlay
        self._fa = FieldAnimation(width, height, field, useCompute,
                options.image)
        if options.draw_field:
            self._fa.drawField = True
        self._t0 = time.time()
        self._fps = 0
        self.options = options
        self.makeGUI(options.gui)

    def makeGUI(self, gui):
        """ Create the GUI context and GlfwRenderer

            Args:
                gui (bool): if True create the GUI context and GlfwRenderer
        """
        if gui:
            imgui.create_context()
            self._renderer = GlfwRenderer(self.window(), True)
            self._opened = True
        else:
            self._renderer = None
            self._opened = False

    def renderScene(self):
        super(GLApp, self).renderScene()
        self._fa.draw()
        self._fps += 1
        if self._opened:
            self._renderer.process_inputs()
            self._opened = userInterface(self._fa, self)
            self._renderer.render(imgui.get_draw_data())
            if not self._opened:
                app.restoreKeyCallback()

        now = time.time()
        if now - self._t0 >= 1:
            if self.options.fps:
                self.setTitle("%s - %s FPS" % (self.title(), self._fps))
            self._fps = 0
            self._t0 = time.time()

    def onKeyboard(self, window, key, scancode, action, mode):
        if key == GLApp.KEY_G and action == GLApp.PRESS:
            # Draw the GUI
            self.makeGUI(not self._opened)

        super(GLApp, self).onKeyboard(window, key, scancode, action, mode)

    def onResize(self, window, width, height):
        gl.glViewport(0, 0, width, height)
        self._fa.setSize(width, height)

    def setField(self, ifield):
        field = createField(CHOICES[ifield])
        self._fa.setField(field)
        self.setTitle('%s' % CHOICES[ifield])
        self.ifield = ifield

#------------------------------------------------------------------------------
def makeParser():
    """ Return the ArgumentParser instance for this script.

        Returns:
            ArgumentParser: instance
    """
    parser = argparse.ArgumentParser(
            description='\nField Animation example. Press "G" to display the'
                    ' GUI while running.',
            add_help=True,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            )

    parser.add_argument('-f', '--draw_field', action='store_true',
            default=False,
            help=("Draw vector field as background image ")
            )
    parser.add_argument('-c', '--choose',
            choices=CHOICES,
            default="epole",
            help=("Choose field to animate ")
            )
    parser.add_argument('-p', '--fps', action='store_true', default=False,
            help=("Count Frames Per Second ")
            )
    parser.add_argument('-u', '--use-fragment', action='store_true',
            default=False, help=("Use fragment instead of compute shader ")
            )
    parser.add_argument('-g', '--gui', action='store_true', default=False,
            help=("Add gui control window ")
            )

    parser.add_argument('-i', '--image', action='store', default=None,
            help=("Load image as background texture")
            )
    return parser

#------------------------------------------------------------------------------
if __name__ == "__main__":
    options = makeParser().parse_args(sys.argv[1:])

    app = GLApp('Field Animation', 800, 800, options)
    app.run()



import OpenGL.GL as gl

class Texture(object):
    def __init__(self, data=None,  width=None, height=None,
                filt=gl.GL_NEAREST, dtype=gl.GL_UNSIGNED_BYTE):
        """ Texture object.
            If data is None an empty texture will be created
        """

        self._data = data

        # format of texture object
        if self._data.ndim > 2 and self._data.shape[2] == 3:
            self._format = gl.GL_RGB
        else:
            self._format = gl.GL_RGBA

        if dtype == gl.GL_FLOAT:
            self._format = gl.GL_R32F
            self._image_format = gl.GL_RED
            self._pname = gl.GL_REPEAT
            self._filter_min = gl.GL_LINEAR
            self._filter_mag = gl.GL_LINEAR
        else:
            self._pname = gl.GL_CLAMP_TO_EDGE
            ## Filtering mode if texture pixels < screen pixels
            self._filter_min = filt
            ## Filtering mode if texture pixels > screen pixels
            self._filter_mag = filt
            self._image_format = self._format

        # Create the Texture
        self._handle = gl.glGenTextures(1)
        if self._data is not None:
            if width is None or height is None:
                try:
                    height, width, bands = self._data.shape
                except ValueError:
                    height, width = self._data.shape

        # Bind texture
        gl.glBindTexture(gl.GL_TEXTURE_2D, self._handle)
        # Avoid banding if row length is odd
        gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)

        # Set Texture wrap and filter modes
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S,
                self._pname);
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T,
                self._pname);
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER,
                self._filter_min);
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER,
                self._filter_mag);
        gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, self._format, width, height,
                0, self._image_format, dtype, self._data)
        # Unbind texture
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0);

    def bind(self, texUnit=0):
        gl.glActiveTexture(gl.GL_TEXTURE0 + texUnit)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self._handle)

    def handle(self):
        return self._handle

    def unbind(self):
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)



"""
FieldAnimation: animate 2D vector fields
========================================

- Field Animation
    Python package to animate 2D vector fields using the power of
    OpenGL graphic cards.

Authors:
    Nicola Creati, ncreati@inogs.it
    Roberto Vidmar, rvidmar@inogs.it
    2018-2020

Istituto Nazionale di Oceanografia e di Geofisica Sperimentale - OGS
https://www.inogs.it

"""
import os
import sys
import pip
from setuptools import setup, find_packages
import distutils.spawn

PKGNAME = "fieldanimation"
SUBVERSION = "2a0"
REQFILE = "requirements.txt"
EXECUTABLES = []
DATAFILES = []
SCRIPTS = ['bin/%s' % sc for sc in EXECUTABLES]
# Retrieve version information
exec(open(os.path.join(PKGNAME, "__version__.py")).read())
version = "%s.%s" % (__version__, SUBVERSION)
# Retrieve requirements
with open(REQFILE) as fp:
    requirements = fp.read()

# Check pip version
if int(pip.__version__.split('.')[0]) < 20:
    # pip upgrade needed!
    print("\n\nWARNING!\n    pip version is too old (%s < 20.0.0)\n"
            % pip.__version__)
    raise SystemExit("\n\n Please run 'pip3 install -U pip' and then"
            " run again this command.\n")

#------------------------------------------------------------------------------
# Check for VIRTUAL_ENV
if 'VIRTUAL_ENV' in os.environ:
    bin_path = os.path.join(os.environ.get('VIRTUAL_ENV'),
                       'lib',
                       'python%d.%d' % sys.version_info[:2],
                       'site-packages', PKGNAME, 'bin')
else:
    bin_path = None

# Avoid name clash for scripts
ok = True
conflicting = []
for ex in EXECUTABLES:
    executable = distutils.spawn.find_executable(ex, bin_path)
    if executable:
        ok = False
        for line in open(executable):
            if ("This script belongs to Python package %s" % PKGNAME in line
                    or (PKGNAME in line
                        and 'EASY-INSTALL-DEV-SCRIPT'in line)):
                # This executable belong to this package
                ok = True
                break
        if not ok:
            conflicting.append(executable)
if not ok:
    raise SystemExit("\nWARNING!\n"
            "Installation will overwrite the following files:\n"
            " --> %s\nPlease resolve conflict before retrying.\n"
            "***Installation aborted***" % conflicting)

#==============================================================================
description = __doc__.split('\n')[1:-1][0]
classifiers = """
Development Status :: 4 - Beta
Intended Audience :: Developers
Intended Audience :: Science/Research
License :: OSI Approved :: MIT License
Operating System :: POSIX
Programming Language :: Python
Topic :: Scientific/Engineering :: Visualization
Topic :: Software Development :: Libraries :: Python Modules
"""

setup(name=PKGNAME,
        version=version,
        install_requires=requirements,
        description=description,
        long_description=open("README.md", "r").read(),
        long_description_content_type='text/markdown',
        classifiers=classifiers.split('\n')[1:-1],
        keywords=[PKGNAME, 'OpenGL', 'Widget', 'Vector field'],
        platforms=['POSIX'],
        license='MIT',
        scripts=SCRIPTS,
        include_package_data=True,
	data_files=DATAFILES,
        url='https://bitbucket.org/bvidmar/fieldanimation',
	download_url='https://bitbucket.org/bvidmar/fieldanimation',
        author='Nicola Creati, Roberto Vidmar',
        author_email='ncreati@inogs.it',
        packages=find_packages(),
	package_data={
            'fieldAnimation': ['glsl/*', ],
            },
        )
